/* Jumpgate
 * Jumpgate is a TCP connection forwarder that provides many enhancements and 
 * improvements over the existing programs that do the same thing.  Only 
 * tested on OpenBSD 2.7, but should run without problems on any Unix-like
 * system.  If you furtherly improve it, please send me the diffs.  As always,
 * published under a BSD style license, see the included LICENSE file.
 *
 * By Patroklos Argyroudis <argp@otenet.gr>.
 */
 
#include "jumpgate.h"

/* global variables, sorry joy */
int rport;			/* remote port */
int log_flag = 0;		/* default is to not log */
struct hostent *host;		/* remote host */
char *logfile;

int
main(int argc, char *argv[])
{
 int lfd, cfd, lport, c;
 const int val = 1;			/* flag for REUSEADDR */
 pid_t chldpid;
 socklen_t clen;
 struct sockaddr_in laddr, caddr;
 int interactive = 0;			/* default is non-interactive mode */
 int backlog = BACKLOG;			/* maximum backlog queue */
 
 rport = lport = 0;
 
 while((c = getopt(argc, argv, "hvil:r:a:f:c:")) != EOF)
 {
  switch(c)
  {
   case 'h':
   	help(argv[0]);
   	exit(EXIT_SUCCESS);
   case 'v':
   	fprintf(stderr, "jumpgate v%s by invisibl <invisible@bsd.gr>\n", VERSION);
   	exit(EXIT_SUCCESS);
   case 'i':
   	interactive = 1;
   	break;
   case 'l':
   	lport = atoi(optarg);
   	if(lport <= 1024 && geteuid() != 0)
   	 xprintf("error: you're trying to bind a well-known port without having root privileges\n");
   	break;
   case 'r':
   	rport = atol(optarg);
   	break;
   case 'a':
   	if(!(host = gethostbyname(optarg)))
   	 xprintf("error resolving %s\n", optarg);
   	break;
   case 'f':
   	log_flag = 1;
   	logfile = optarg;
   	break;
   case 'c':
   	backlog = atoi(optarg);
   	break;
   default:
   	usage(argv[0]);
   	exit(EXIT_SUCCESS);
  }
 }
 
 /* check the users input */
 if(lport == 0 && rport == 0 && host == NULL && interactive == 0)
 {
  usage(argv[0]);
  exit(EXIT_SUCCESS);
 }
 
 if(lport == 0)
  xprintf("error: you specified no local port\n");

 if(rport == 0 && interactive == 0)
  xprintf("error: you specified no remote port\n");
  
 if(host == NULL && interactive == 0)
  xprintf("error: you specified no host or IP address\n");
   
 if(daemonize(".") == -1)
  xprintf("fork error\n");
 
 if((lfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
  xprintf("socket error\n");
 
 bzero(&laddr, sizeof(laddr));
 laddr.sin_family = AF_INET;
 laddr.sin_addr.s_addr = htonl(INADDR_ANY);
 laddr.sin_port = htons(lport);
 
 /* if you use linux beware of the SO_REUSEADDR vulnerability,
  * although I think new linux versions fix this problem
  */
 if(setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val)) < 0)
  xprintf("setsockopt error\n");
 
 if(bind(lfd, (struct sockaddr *)&laddr, sizeof(laddr)) < 0)
  xprintf("bind error\n");

 if(listen(lfd, backlog) < 0)
  xprintf("listen error\n");
 
 if(signal(SIGCHLD, chld_reap) == SIG_ERR)
  xprintf("signal error\n");
 
 for(; ;)
 {
  clen = sizeof(caddr);
  
  if((cfd = accept(lfd, (struct sockaddr *)&caddr, &clen)) < 0)
  {
   if(errno == EINTR)
    continue;
   else
    xprintf("accept error\n");
  }
  
  if((chldpid = fork()) == -1)
   xprintf("fork error\n");
  
  if(chldpid == 0)
  {
   if(close(lfd) == -1)
    xprintf("close error\n");
   
   if(interactive == 1)
    query(cfd);
    
   jumpgate(cfd, rport, host);
   exit(EXIT_SUCCESS);
  }
  
  if(close(cfd) == -1)
   xprintf("close error\n");
 }
 
 return(EXIT_SUCCESS);
}

void
usage(char *name)
{
 fprintf(stderr, "usage: %s [-hvi] [-l <local port>] [-r <remote port>]\n", name);
 fprintf(stderr, "[-a <remote host or IP address>] [-f <filename>] [-c <number of connections>]\n");
}

void
help(char *name)
{
 fprintf(stderr, "jumpgate v%s by invisibl <invisible@bsd.gr>\n\n", VERSION);
 fprintf(stderr, "usage: %s [-hvi] [-l <local port>] [-r <remote port>]\n", name);
 fprintf(stderr, "[-a <remote host or IP address>] [-f <filename>] [-c <number of connections>]\n\n");
 fprintf(stderr, "-h		this help message\n");
 fprintf(stderr, "-v		display version number and exit\n");
 fprintf(stderr, "-i		interactively ask the user where to forward\n");
 fprintf(stderr, "-l		specify the local listening port\n");
 fprintf(stderr, "-r		specify the remote port to forward the connection\n");
 fprintf(stderr, "-a		specify the host to forward the connection\n"); 
 fprintf(stderr, "-f		specify the filename to log the connection\n");
 fprintf(stderr, "-c		specify the maximum number of the connections backlog\n");
}
 
void
xprintf(char *msg, ...)
{
 char buffer[SMALLBUFF];
 va_list args;
 
 va_start(args, msg);
 vsnprintf(buffer, SMALLBUFF, msg, args);
 
 fprintf(stderr, "%s", buffer);
 va_end(args);
 exit(EXIT_FAILURE);
}

void
chld_reap(int signo)
{
 pid_t pid;
 int stat;
  
 while((pid = waitpid(-1, &stat, WNOHANG)) > 0)
  /* do nothing */ ;
       
 return;
}

int
daemonize(const char *path)
{
 pid_t pid;
 
 if((signal(SIGTTOU, SIG_IGN)) == SIG_ERR)
  xprintf("signal error\n");
        
 if((signal(SIGTTIN, SIG_IGN)) == SIG_ERR)
  xprintf("signal error\n");
               
 if((signal(SIGTSTP, SIG_IGN)) == SIG_ERR)
  xprintf("signal error\n");
                      
 if((signal(SIGHUP, SIG_IGN)) == SIG_ERR)
  xprintf("signal error\n");
                             
 if((pid = fork()) < 0)
  return(-1);
 else if(pid != 0)
  exit(0);
                                   
 if(setsid() == -1)
  xprintf("setsid error\n");
                                          
 if(chdir(path) == -1)
  xprintf("chdir error\n");
                                                 
 umask(0);
 return(0);
}

void
query(int fd)
{
 u_char p[MAXLINE];
 u_char hname[MAXLINE];
 u_char msg[MAXLINE];
 u_char reply[MAXLINE];
 
 snprintf(msg, sizeof(msg), "<hostname or IP address> <port> (e.g.: bsd.gr 80): ");
 if((write(fd, msg, strlen(msg))) == -1)
  xprintf("write error\n");
 
 if((read(fd, reply, (sizeof(reply) - 1))) == -1)
  xprintf("read error\n");
  
 sscanf(reply, "%s %s", hname, p);
 rport = atol(p);
 if(!(host = gethostbyname(hname)))
 {
  snprintf(msg, sizeof(msg), "error resolving %s\n", hname);
  if((write(fd, msg, strlen(msg))) == -1)
   xprintf("write error\n");
   
  exit(EXIT_FAILURE);
 }
 
 return;
} 

void
jumpgate(int fdc, int port, struct hostent *h)
{
 FILE *fp;
 int n, ofd;
 char buf[BUFFSIZE];
 fd_set fdsr, fdse;
 struct sockaddr_in oaddr;
 
 if(log_flag == 1)
 {
  if((fp = fopen(logfile, "a")) == NULL)
   xprintf("fopen error\n");
 }
 
 if((ofd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
  xprintf("socket error\n");
 
 oaddr.sin_family = h->h_addrtype;
 oaddr.sin_port = htons(port);
 memcpy(&oaddr.sin_addr, h->h_addr, h->h_length);
 
 if((connect(ofd, (struct sockaddr *)&oaddr, sizeof(oaddr))) < 0)
  xprintf("connect error\n");
 
 if(log_flag == 1)
  fprintf(fp, "--[ host: %s port: %d ]--\n", h->h_name, port);
 
 for(; ;)
 {
  FD_ZERO(&fdsr);
  FD_ZERO(&fdse);
  FD_SET(fdc, &fdsr);
  FD_SET(fdc, &fdse);
  FD_SET(ofd, &fdsr);
  FD_SET(ofd, &fdse);
  
  if(select(20, &fdsr, NULL, &fdse, NULL) == -1)
  {
   shutdown(ofd, 2);
   close(ofd);
   return;
  }
  
  if(FD_ISSET(fdc, &fdsr) || FD_ISSET(fdc, &fdse))
  {
   if((n = read(fdc, buf, BUFFSIZE)) <= 0)
   {
    shutdown(ofd, 2);
    close(ofd);
    return;
   }
   else if(log_flag == 1)
    fprintf(fp, "%s\n", buf);
   
   if((write(ofd, buf, n)) <= 0)
   {
    shutdown(ofd, 2);
    close(ofd);
    return;
   }
  } 
  else if(FD_ISSET(ofd, &fdsr) || FD_ISSET(ofd, &fdse))
  {
   if((n = read(ofd, buf, BUFFSIZE)) < 0)
   {
    shutdown(ofd, 2);
    close(ofd);
    return;
   }
   
   if((write(fdc, buf, n)) <= 0)
   {
    shutdown(ofd, 2);
    close(ofd);
    return;
   }
  }
 }
 return;
} 

/* EOF */
